from __future__ import annotations

import math
from dataclasses import dataclass
from typing import Generic, TypeVar
from uuid import UUID, uuid4

from cs.structures.heap.heap import Heap
from cs.util import Comparable, dfield, formatter

T = TypeVar("T", bound=Comparable)


@dataclass(order=True)
class Entry(Generic[T]):
    """
    Hold an entry in the heap.
    In order for all of the Binomial heap operations to complete in O(1),
    clients need to have O(1) access to any element in the heap. We make
    this work by having each insertion operation produce a handle to the
    node in the tree. In actuality, this handle is the node itself.

    Priority is the first parameter because the dataclass orders Entry as a
    tuple (priority, value)
    """

    priority: float
    value: T
    child: Entry[T] | None = dfield(None, repr=True)
    right: Entry[T] | None = dfield(None, repr=True)

    def __repr__(self) -> str:
        return str(formatter.pformat(self))

    def __len__(self) -> int:
        """
        Recursively calculate length. Since we never use this function except
        in tests, this is simpler than keeping track of each node's size.
        """
        if self.child is None:
            return 1 if self.right is None else len(self.right) + 1
        if self.right is None:
            return len(self.child) + 1
        return len(self.child) + len(self.right) + 1


@dataclass(init=False)
class BinomialHeap(Heap[T]):
    """
    Binomial Heap implementation using a list of trees, represented by Entry objects.

    Note that the input type should be bounded by Comparable in order for the
    implementation to type-check. Otherwise, the heap should still work so long as
    there are no equal priorities. If you want to use a non-comparable type with this
    class, use a wrapper class with a custom comparator or only use unique priorities.
    """

    trees: list[Entry[T] | None]

    def __init__(self, *, allow_duplicates: bool = False) -> None:
        # Trees in ascending order: 1, 2, 4, 8, ..., etc.
        self.trees: list[Entry[T] | None] = []

        # Mapping from element to corresponding entry.
        # Should not introduce any asymptotic change in memory overhead.
        self.elem_to_entry: dict[UUID | T, Entry[T]] = {}

        # Cached size of the heap, so we don't have to recompute this explicitly.
        self.size = 0

        # Whether to allow duplicate key entries, which means using UUIDs for
        # elem_to_entry's keys instead.
        self.allow_duplicates = allow_duplicates

    def __bool__(self) -> bool:
        return bool(self.trees)

    def __len__(self) -> int:
        return self.size

    def __contains__(self, item: T) -> bool:
        if self.allow_duplicates:
            # TODO
            raise NotImplementedError
        return item in self.elem_to_entry

    def __getitem__(self, value: T | UUID) -> Entry[T]:
        """Gets the correct Entry object from the given value or UUID."""
        if self.allow_duplicates and not isinstance(value, UUID):
            raise RuntimeError(
                "You must pass in a valid UUID or set allow_duplicates = False."
            )
        if not self.allow_duplicates and isinstance(value, UUID):
            raise RuntimeError("You must pass in a value of type T, not a UUID.")

        if value not in self.elem_to_entry:
            raise KeyError(
                f"Invalid {'UUID' if self.allow_duplicates else 'key'}: {value}"
            )

        return self.elem_to_entry[value]

    def __or__(self, other: object) -> BinomialHeap[T]:
        # TODO - this is a destructive operation
        if isinstance(other, BinomialHeap):
            result = BinomialHeap[T]()
            self.merge(other)
            result.merge(self)
            return result
        raise NotImplementedError

    def __ior__(self, other: object) -> None:
        if isinstance(other, BinomialHeap):
            self.merge(other)
        raise NotImplementedError

    @staticmethod
    def merge_lists(
        one: list[Entry[T] | None], two: list[Entry[T] | None]
    ) -> list[Entry[T] | None]:
        """
        Merging two binomial heaps is similar to adding two binary numbers. We proceed
        from the "least-significant tree" to the "most-significant tree", merging the
        two trees and storing the result either back in the same slot (if no trees were
        added) or in a carry register to be used in the next computation. This next
        variable declaration contains the carry.

        @param one A reference to one of the two deques.
        @param two A reference to the other of the two deques.
        @return A reference to the smallest element of the resulting list.
        """

        def _merge_trees(one: Entry[T], two: Entry[T]) -> Entry[T]:
            """Merge 2 trees."""
            if two < one:
                one, two = two, one
            two.right = one.child
            one.child = two
            return one

        max_order = max(len(one), len(two))
        lhs = one + [None] * (max_order - len(one))
        rhs = two + [None] * (max_order - len(two))

        result: list[Entry[T] | None] = []
        carry: Entry[T] | None = None
        for order in range(max_order):
            # There are eight possible combinations of the None-ity of the carry,
            # lhs, and rhs trees. To make the logic simpler, we'll add them all to
            # a temporary buffer and proceed from there.
            trees: list[Entry[T]] = []
            if carry is not None:
                trees.append(carry)
            if (left := lhs[order]) is not None:
                trees.append(left)
            if (right := rhs[order]) is not None:
                trees.append(right)

            # Case one: both trees and the carry are None. Then the result of
            # this step is None and the carry should be cleared.
            if not trees:
                result.append(None)
                carry = None
            # Case two: There's exactly one tree. Then the result of this
            # step is that tree and the carry is cleared.
            elif len(trees) == 1:
                result.append(trees[0])
                carry = None
            # Case three: There's exactly two trees. Then the result of this
            # operation is None and the carry will be set to the merge of those trees.
            elif len(trees) == 2:
                result.append(None)
                carry = _merge_trees(trees[0], trees[1])
            # Case four: There's exactly three trees. Then we'll arbitrarily
            # store one of them in the current slot, then put the merge of the
            # other two into the carry.
            else:
                result.append(trees[0])
                carry = _merge_trees(trees[1], trees[2])

        # Finally, if the carry is set, append it to the result.
        if carry is not None:
            result.append(carry)
        return result

    @staticmethod
    def _check_priority(priority: float) -> None:
        """
        Given a user-specified priority, check whether it's a valid double
        and throw a ValueError otherwise.

        @param priority The user's specified priority.
        @raises ValueError if it is not valid.
        """
        if math.isnan(priority):
            raise ValueError(f"Priority {priority} is invalid.")

    def enqueue(self, value: T, priority: float = 0) -> T | UUID:
        """
        Insert an element into the Binomial heap with the specified priority.

        Its priority must be a valid double, so you cannot set the priority to NaN.

        @param value The value to insert.
        @param priority Its priority, which must be valid.
        @return An Entry representing that element in the tree.
        """
        self._check_priority(priority)
        if not self.allow_duplicates and value in self.elem_to_entry:
            raise KeyError(
                f"Duplicate key detected: {value}. "
                "Use allow_duplicates = True to allow duplicate entries using UUIDs."
            )

        # Create the entry object, which is a circularly-linked list of length one.
        result = Entry(priority, value)

        # Merge this singleton list with the tree list.
        self.trees = self.merge_lists(self.trees, [result])
        self.size += 1

        key: T | UUID = uuid4() if self.allow_duplicates else value
        self.elem_to_entry[key] = result
        return key

    def peek(self) -> tuple[T, float]:
        """
        Return an Entry object corresponding to the minimum element of the heap.

        Raise an IndexError if the heap is empty.

        @return The smallest element of the heap.
        @raises IndexError If the heap is empty.
        """
        if not self.trees:
            raise IndexError("Heap is empty.")
        top = min(entry for entry in self.trees if entry is not None)
        return top.value, top.priority

    def dequeue(self) -> tuple[T, float]:
        """
        Dequeue and return the minimum element of the Binomial heap.

        If the heap is empty, this throws an IndexError.

        @return The smallest element of the Binomial heap.
        @raises IndexError if the heap is empty.
        """
        if not self.trees:
            raise IndexError("Heap is empty.")

        result = min(entry for entry in self.trees if entry is not None)
        index = self.trees.index(result)
        children: list[Entry[T] | None] = []
        entry = result.child
        while entry is not None:
            children.append(entry)
            next_ = entry.right
            entry.right = None
            entry = next_
        del self.trees[index]

        # Shrink forest size if we just got rid of the largest tree size.
        if self.trees and self.trees[-1] is None:
            del self.trees[-1]

        # These children were added in reverse order because as they're
        # merged, higher-order trees have lower-order trees as children
        # but the trees are stored in ascending orders in the vectors.
        # Therefore, we need to reverse the list.
        children.reverse()
        self.trees = self.merge_lists(self.trees, children)
        self.size -= 1
        if not self.allow_duplicates:
            del self.elem_to_entry[result.value]
        return result.value, result.priority

    def merge(self, other: Heap[T]) -> None:
        """
        Merge 2 Binomial heaps.

        Given two Binomial heaps, obtain a Binomial heap that contains
        all of the elements of the two heaps in place.

        @param self The first Binomial heap to merge.
        @param other The second Binomial heap to merge.
        """
        if not isinstance(other, BinomialHeap):
            raise TypeError("Heap types must match when merging.")
        if not self.allow_duplicates and set(self.elem_to_entry) & set(
            other.elem_to_entry
        ):
            raise RuntimeError(
                "You must pass in two unoverlapping heaps or set "
                "allow_duplicates = True on both heaps."
            )

        # Clear out the rhs and assign the lhs the value of result.
        self.trees = self.merge_lists(self.trees, other.trees)

        # The size of the new heap is the sum of the sizes of the input heaps.
        self.size += other.size
        self.allow_duplicates = self.allow_duplicates or other.allow_duplicates
        self.elem_to_entry |= other.elem_to_entry



from dataclasses import dataclass, field
from typing import Generic, TypeVar

T = TypeVar("T")


@dataclass(slots=True)
class Stack(Generic[T]):
    """You should probably use the Python built-in List instead."""

    _stack: list[T] = field(default_factory=list)

    def __bool__(self) -> bool:
        return bool(self._stack)

    def __len__(self) -> int:
        return len(self._stack)

    def __contains__(self, item: T) -> bool:
        return item in self._stack

    def push(self, item: T) -> None:
        self._stack.append(item)

    def pop(self) -> T:
        if not self._stack:
            raise IndexError
        return self._stack.pop()

    def peek(self) -> T:
        if not self._stack:
            raise IndexError
        return self._stack[-1]



from .molecular_chemistry import (
    molarity_to_normality,
    moles_to_pressure,
    moles_to_volume,
    pressure_and_volume_to_temperature,
)
from .roman_to_integer import roman_to_int
from .si_units import Binary_Unit, SI_Unit, convert_binary_prefix, convert_si_prefix
from .temperature import (
    celsius_to_fahrenheit,
    celsius_to_kelvin,
    celsius_to_rankine,
    fahrenheit_to_celsius,
    fahrenheit_to_kelvin,
    fahrenheit_to_rankine,
    kelvin_to_celsius,
    kelvin_to_fahrenheit,
    kelvin_to_rankine,
    rankine_to_celsius,
    rankine_to_fahrenheit,
    rankine_to_kelvin,
)

__all__ = (
    "Binary_Unit",
    "SI_Unit",
    "celsius_to_fahrenheit",
    "celsius_to_kelvin",
    "celsius_to_rankine",
    "convert_binary_prefix",
    "convert_si_prefix",
    "fahrenheit_to_celsius",
    "fahrenheit_to_kelvin",
    "fahrenheit_to_rankine",
    "kelvin_to_celsius",
    "kelvin_to_fahrenheit",
    "kelvin_to_rankine",
    "molarity_to_normality",
    "moles_to_pressure",
    "moles_to_volume",
    "pressure_and_volume_to_temperature",
    "rankine_to_celsius",
    "rankine_to_fahrenheit",
    "rankine_to_kelvin",
    "roman_to_int",
)



from __future__ import annotations

from dataclasses import dataclass, field

from cs.util import formatter


@dataclass
class Trie:
    """
    Does not use a TrieNode struct because the root node can contain
    several children nodes, thus making the internal logic more complex.
    Instead, sets char to the empty string for the head of the Trie.
    """

    char: str = ""
    is_leaf: bool = False
    size: int = 0
    children: dict[str, Trie] = field(default_factory=dict)

    def __str__(self) -> str:
        return str(formatter.pformat(self))

    def __len__(self) -> int:
        return self.size

    def __bool__(self) -> bool:
        return bool(self.children)

    def __contains__(self, prefix: str) -> bool:
        """
        Tries to find word in a Trie
        :param word: word to look for
        :return: Returns True if word is found, False otherwise
        """
        trie = self
        for ch in prefix:
            if ch not in trie.children:
                return False
            trie = trie.children[ch]
        return trie.is_leaf

    def insert(self, text: str) -> None:
        """
        Inserts a word into the Trie
        :param word: word to be inserted
        :return: None
        """
        trie = self
        for ch in text:
            trie.size += 1
            if ch not in trie.children:
                trie.children[ch] = Trie(ch)
            trie = trie.children[ch]
        trie.is_leaf = True

    def remove(self, word: str) -> None:
        """
        Deletes a word in a Trie
        :param word: word to delete
        :return: None
        """

        def _remove(curr: Trie, word: str) -> bool:
            # If word is empty, attempt to set the word to not a leaf.
            # If the word has no other children,
            # return False so that we can delete above keys.
            curr.size -= 1
            if not word:
                if not curr.is_leaf:
                    return False
                curr.is_leaf = False
                return not curr.children

            ch = word[0]
            if ch not in curr.children:
                return False

            should_delete_curr = _remove(curr.children[ch], word[1:])
            if should_delete_curr:
                del curr.children[ch]
                return not curr.children
            return should_delete_curr

        if word not in self:
            raise KeyError(f"Trie does not contain key: {word}")

        _ = _remove(self, word)



from __future__ import annotations

from collections.abc import Callable, Iterable, Iterator, KeysView, Mapping, Sequence
from dataclasses import dataclass
from typing import Any, Generic, TypeVar

from cs.util import Comparable, formatter

V = TypeVar("V", bound=Comparable)


@dataclass(init=False)
class Graph(Generic[V]):
    """
    Use a Dict of Dicts to avoid storing nodes with no edges, as well as
    provide instant lookup for nodes and their neighbors.

    Designed for extensibility - a user can easily add/extend a
    custom Node or Edge class and include it in the type checking system.
    """

    INFINITY = float("inf")
    INT_INFINITY = 1 << 20
    _graph: dict[V, dict[V, Edge[V]]]
    is_directed: bool

    def __init__(
        self,
        graph: dict[V, Any] | None = None,
        *,
        is_directed: bool = True,
        weight: float = 1,
        **kwargs: Any,
    ) -> None:
        """Default constructor assumes an adjacency list representation."""
        self.is_directed = is_directed
        self._graph = {}
        if graph is not None:
            for u in graph:
                self.add_node(u)
            for u in graph:
                for v in graph[u]:
                    if v not in self._graph:
                        self.add_node(v)
                    if isinstance(graph[u], dict):
                        edge = graph[u][v]
                        if isinstance(edge, Edge):
                            self.add_edge(**edge)
                        elif isinstance(edge, dict):
                            self.add_edge(u, v, **edge)
                        elif isinstance(edge, (int, float)):
                            self.add_edge(u, v, weight=edge, **kwargs)
                        else:
                            raise TypeError(f"{edge} is not a supported Edge type.")
                    elif isinstance(graph[u], (list, tuple, set, frozenset)):
                        # Values are some other collection; only contains node names.
                        # Use default weight parameter.
                        self.add_edge(u, v, weight)
                    else:
                        raise TypeError(f"{graph[u]} is not a supported Edge mapping.")

    def __str__(self) -> str:
        return str(formatter.pformat(self._graph))

    def __len__(self) -> int:
        return len(self._graph)

    def __bool__(self) -> bool:
        return bool(self._graph)

    def __contains__(self, v: V) -> bool:
        return v in self._graph

    def __getitem__(self, v: V) -> dict[V, Edge[V]]:
        self.verify_nodes_exist(v)
        return self._graph[v]

    def __iter__(self) -> Iterator[V]:
        yield from self._graph

    @property
    def nodes(self) -> KeysView[V]:
        return self._graph.keys()

    @property
    def edges(self) -> tuple[Edge[V], ...]:
        return tuple(self._graph[u][v] for u in self._graph for v in self._graph[u])

    @classmethod
    def from_graph(
        cls,
        graph: Graph[V],
        *,
        is_directed: bool = True,
        node_fn: Callable[[V], Any] = lambda x: x,
        edge_fn: Callable[[Edge[V]], Edge[V]] = lambda x: x,
    ) -> Graph[Any]:
        """
        This function is used to make a copy of a graph, or to apply transformations
        and return a new version of the graph.
        We manually copy the contents of the graph in order to avoid sharing the
        same references to neighbor dicts.
        """
        new_graph = Graph[Any](is_directed=is_directed)
        for u in graph:
            new_graph.add_node(node_fn(u))
        for u in graph:
            for v in graph[u]:
                new_u, new_v = node_fn(u), node_fn(v)
                edge = edge_fn(graph[u][v])
                new_graph.add_edge(new_u, new_v, edge.weight, **edge.kwargs)
        return new_graph

    @classmethod
    def from_edgelist(
        cls, edge_list: Iterable[Edge[V]], *, is_directed: bool = True
    ) -> Graph[V]:
        graph = Graph[V](is_directed=is_directed)
        for edge in edge_list:
            if edge.start not in graph:
                graph.add_node(edge.start)
            if edge.end not in graph:
                graph.add_node(edge.end)
            graph.add_edge(**edge)
        return graph

    @classmethod
    def from_matrix(
        cls, matrix: Sequence[Sequence[float]], *, zero_is_no_edge: bool = True
    ) -> Graph[int]:
        """By default, treat edges with weight 0 as non-existent edges."""
        is_directed = False
        n = len(matrix)
        graph: dict[int, dict[int, float]] = {i: {} for i in range(n)}
        for i in range(n):
            for j in range(n):
                edge_data = matrix[i][j]
                # If matrix is not symmetric, graph is directed
                if not is_directed and i < j and edge_data != matrix[j][i]:
                    is_directed = True
                # Only add edges with nonzero edges.
                if edge_data != Graph.INFINITY and not (
                    zero_is_no_edge and edge_data == 0
                ):
                    graph[i][j] = edge_data
        return Graph(graph, is_directed=is_directed)

    @staticmethod
    def is_undirected(graph: Any) -> bool:
        """
        Returns whether the graph can be interpreted as undirected - in other words,
        all forward edges have an accompanying reverse edge.
        """
        if not graph:
            return False
        for u in graph:
            for v in graph[u]:
                if u not in graph[v]:
                    return False
        return True

    def to_graphviz(self, img_format: str = "png") -> None:
        from graphviz import Graph as GraphViz  # type: ignore[import]

        dot = GraphViz(format=img_format)
        for node in self.nodes:
            dot.node(str(node))
        for edge in self.edges:
            dot.edge(str(edge.start), str(edge.end))
        dot.render()

    def to_matrix(self, *, zero_is_no_edge: bool = True) -> list[list[float]]:
        """By default, outputs non-existent edges as having weight 0."""
        nodes = sorted(self._graph)
        graph = [
            [0 if zero_is_no_edge else Graph.INFINITY for _ in nodes] for _ in nodes
        ]
        for i, u in enumerate(nodes):
            for j, v in enumerate(nodes):
                if v in self._graph[u]:
                    graph[i][j] = self._graph[u][v].weight
        return graph

    def verify_nodes_exist(self, *v_ids: V) -> None:
        """Checks existence of provided nodes."""
        for v in v_ids:
            if v not in self._graph:
                raise KeyError(f"Node not found: {v}")

    def adj(self, v: V) -> KeysView[V]:
        """Used for getting the neighbors of a node in a list-like form."""
        self.verify_nodes_exist(v)
        return self._graph[v].keys()

    def degree(self, v: V) -> int:
        """
        Returns the total number of edges going in or out of a node.
        For undirected graphs, counts each edge only once.
        """
        self.verify_nodes_exist(v)
        if self.is_directed:
            return self.out_degree(v) + self.in_degree(v)
        return len(self._graph[v])

    def out_degree(self, v: V) -> int:
        if not self.is_directed:
            raise NotImplementedError("Graph is undirected; use degree() instead.")
        self.verify_nodes_exist(v)
        return len(self._graph[v])

    def in_degree(self, v: V) -> int:
        """Iterate over neighbors to see whether any reference the current node."""
        if not self.is_directed:
            raise NotImplementedError("Graph is undirected; use degree() instead.")
        self.verify_nodes_exist(v)
        return sum(v in self._graph[node] and v != node for node in self._graph)

    def add_node(self, v: V) -> None:
        """You cannot add the same node twice."""
        if v in self._graph:
            raise KeyError(f"Node already exists: {v}")
        self._graph[v] = {}

    def add_edge(self, start: V, end: V, weight: float = 1, **kwargs: Any) -> None:
        """
        For directed graphs, connects the edge start -> end.
        For undirected graphs, also connects the edge end -> start.
        Replaces the edge if it already exists.

        Note that the function variable names must match the Edge class constructor.
        """
        self.verify_nodes_exist(start, end)
        self._graph[start][end] = Edge(start, end, weight, **kwargs)
        if not self.is_directed:
            # pylint: disable=arguments-out-of-order
            self._graph[end][start] = Edge(end, start, weight, **kwargs)

    def has_edge(self, start: V, end: V) -> bool:
        return start in self._graph and end in self._graph[start]

    def remove_node(self, v: V) -> None:
        """
        Removes all of the edges associated with the node v too.
        """
        self.verify_nodes_exist(v)
        if self.is_directed:
            for node in self._graph:
                # Make a list copy to avoid removing-while-iterating errors.
                for neighbor in list(self._graph[node]):
                    if v in (node, neighbor):
                        del self._graph[node][neighbor]
            del self._graph[v]
        else:
            removed_node_dict = self._graph.pop(v)
            for neighbor in removed_node_dict:
                if v in self._graph[neighbor]:
                    del self._graph[neighbor][v]

    def remove_edge(self, start: V, end: V) -> None:
        self.verify_nodes_exist(start, end)
        if end in self._graph[start]:
            del self._graph[start][end]
            if not self.is_directed:
                del self._graph[end][start]

    def is_bipartite(self) -> bool:
        """
        Check whether Graph is bipartite using DFS.
        Should not be a property because the calculation changes and
        this should not be thought of as an easily accessible attribute.

        The algorithm works by exploring nodes in a DFS fashion and marking the parity
        of each node. As an optimization, we combine this into the depth-first search
        code. If we ever explore an arc where both nodes are known to have the same
        parity, we have detected an odd cycle and return False. Since the DFS considers
        each arc in the graph twice (once in each direction), we're guaranteed that we
        will locate such an edge if it exists.

        Runtime: O(|V| + |E|)
        """
        visited = set()
        color: dict[V, bool] = {}

        def dfs(v: V, curr_color: bool) -> None:
            visited.add(v)
            color[v] = curr_color
            for u in self._graph[v]:
                if u not in visited:
                    dfs(u, not curr_color)

        for node in self._graph:
            if node not in visited:
                dfs(node, True)
        for i in self._graph:
            for j in self._graph[i]:
                if color[i] == color[j]:
                    return False
        return True


@dataclass(init=False, repr=False, order=True)
class Edge(Generic[V], Mapping[str, Any]):
    """
    The edge class that stores edge data.
    Edges are given sort order using start, end, and weight.
    """

    start: V
    end: V
    weight: float

    def __init__(self, start: V, end: V, weight: float = 1, **kwargs: Any):
        self.start = start
        self.end = end
        self.weight = weight
        self.kwargs = kwargs

    def __len__(self) -> int:
        return 3 + len(self.kwargs)

    def __iter__(self) -> Iterator[Any]:
        yield from ("start", "end", "weight")
        yield from self.kwargs

    def __getitem__(self, attr: str) -> Any:
        if attr == "start":
            return self.start
        if attr == "end":
            return self.end
        if attr == "weight":
            return self.weight
        return self.kwargs[attr]

    def __setitem__(self, attr: str, value: Any) -> None:
        if attr == "start":
            self.start = value
        if attr == "end":
            self.end = value
        if attr == "weight":
            self.weight = value
        self.kwargs[attr] = value

    def __repr__(self) -> str:
        result = str(formatter.pformat(self))[:-1]
        for key, kwarg in self.kwargs.items():
            result += f", {key}={kwarg}"
        return result + ")"

    def __hash__(self) -> int:
        return hash((self.start, self.end))


@dataclass(order=True, slots=True)
class Node(Generic[V]):
    """An example node class that stores node data."""

    data: V

    def __hash__(self) -> int:
        return hash(self.data)


class DirectedGraph(Generic[V], Graph[V]):
    def __init__(self, graph: dict[V, Any] | None = None) -> None:
        super().__init__(graph)


class UndirectedGraph(Generic[V], Graph[V]):
    def __init__(self, graph: dict[V, Any] | None = None) -> None:
        super().__init__(graph, is_directed=False)



import hashlib
import math
from collections.abc import Iterator
from dataclasses import dataclass
from typing import Generic, TypeVar

from bitarray import bitarray

T = TypeVar("T")


@dataclass
class BloomFilter(Generic[T]):
    """
    Implementation of a Bloom filter. An instance is initialized by its capacity `n`
    and error rate `p`. The capacity `n` is how many elements can be stored while
    maintaining no more than `p` false positives.

    Code adapted from @ilanschnell's bitarray project.
    """

    n: int
    p: float = 0.01

    def __post_init__(self) -> None:
        if not 0 < self.p < 1:
            raise ValueError("Invalid value of p.")
        self.k = math.ceil(-math.log2(self.p))  # number of hash functions
        self.m = math.ceil(-self.n * math.log2(self.p) / math.log(2))  # size of array
        self.array = bitarray(self.m)
        self.array.setall(0)
        self.size = 0

    def __len__(self) -> int:
        return self.size

    def __bool__(self) -> bool:
        return any(self.array)

    def __contains__(self, key: T) -> bool:
        return all(self.array[i] for i in self._hashes(key))

    def __str__(self) -> str:
        return f"{self.__class__.__qualname__}({self.array.to01()})"

    def calculate_p(self) -> float:
        """
        Calculate the actual false positive error rate `p` from the number of hashes `k`
        and the size if the bitarray `m`. This is slightly different from the given `p`,
        because the integer value of `k` is being used.
        """
        return pow(1 - math.exp(-self.k * self.n / self.m), self.k)

    def approx_items(self) -> int:
        """Return the approximate number of items in the Bloom filter."""
        count = self.array.count()
        return 0 if count == 0 else int(-self.m / self.k * math.log(1 - count / self.m))

    def add(self, key: T) -> None:
        self.size += 1
        for i in self._hashes(key):
            self.array[i] = 1

    def _hashes(self, key: T) -> Iterator[int]:
        """
        Generate k different hashes, each of which maps a key to one of
        the m array positions with a uniform random distribution.
        """
        h = hashlib.new("md5")
        h.update(str(key).encode())
        x = int(h.hexdigest(), 16)
        for _ in range(self.k):
            if x < 1024 * self.m:
                h.update(b"x")
                x = int(h.hexdigest(), 16)
            x, y = divmod(x, self.m)
            yield y



from __future__ import annotations

from collections.abc import Iterator
from dataclasses import dataclass
from typing import TypeVar

from cs.structures.tree.tree import Tree, TreeNode
from cs.util import Comparable, dfield

T = TypeVar("T", bound=Comparable)


@dataclass(order=True, repr=False, slots=True)
class BinaryTreeNode(TreeNode[T]):
    left: BinaryTreeNode[T] | None = None
    right: BinaryTreeNode[T] | None = None
    parent: BinaryTreeNode[T] | None = dfield(None)

    @property
    def grandparent(self) -> BinaryTreeNode[T] | None:
        """Get the current node's grandparent, or None if it doesn't exist."""
        return None if self.parent is None else self.parent.parent

    @property
    def sibling(self) -> BinaryTreeNode[T] | None:
        """Get the current node's sibling, or None if it doesn't exist."""
        if self.parent is None:
            return None
        return self.parent.right if self.parent.left is self else self.parent.left

    def is_root(self) -> bool:
        """Returns true iff this node is the root of the tree."""
        return self.parent is None

    def is_left(self) -> bool:
        """Returns true iff this node is the left child of its parent."""
        return self.parent is not None and self.parent.left is self

    def is_right(self) -> bool:
        """Returns true iff this node is the right child of its parent."""
        return self.parent is not None and self.parent.right is self


@dataclass
class BinarySearchTree(Tree[T]):
    """
    BinarySearchTrees contain nodes in sorted order.

    We separate the BinarySearchTree with the TreeNode class to allow the root
    of the tree to be None, which allows this implementation to type-check.
    """

    root: BinaryTreeNode[T] | None = None
    size: int = 0

    def __iter__(self) -> Iterator[BinaryTreeNode[T]]:
        """Performs an in-order traversal over the TreeNodes of the Tree."""

        def _iter(node: BinaryTreeNode[T] | None) -> Iterator[BinaryTreeNode[T]]:
            if node is not None:
                yield from _iter(node.left)
                yield node
                yield from _iter(node.right)

        return _iter(self.root)

    def __str__(self) -> str:
        from cs.structures.tree.draw_tree import draw_tree

        return draw_tree(self.root)

    @staticmethod
    def depth(tree: BinaryTreeNode[T] | None) -> int:
        if tree is None:
            return 0
        return 1 + max(
            BinarySearchTree.depth(tree.left), BinarySearchTree.depth(tree.right)
        )

    def height(self) -> int:
        return self.depth(self.root)

    def is_balanced(self) -> bool:
        if self.root is None:
            raise Exception("Binary search tree is empty")
        return self.depth(self.root.left) == self.depth(self.root.right)

    def search(self, data: T) -> BinaryTreeNode[T] | None:
        """Searches a node in the tree."""

        def _search(node: BinaryTreeNode[T] | None) -> BinaryTreeNode[T] | None:
            if node is None:
                return None
            if node.data == data:
                node.hits += 1
                return node
            return _search(node.left) if data < node.data else _search(node.right)

        return _search(self.root)

    def insert(self, data: T) -> None:
        """Puts a new node in the tree."""

        def _insert(
            node: BinaryTreeNode[T] | None, parent: BinaryTreeNode[T] | None = None
        ) -> BinaryTreeNode[T]:
            if node is None:
                node = BinaryTreeNode(data, parent=parent)
                return node
            if data == node.data:
                node.count += 1
            elif data < node.data:
                node.left = _insert(node.left, node)
            else:
                node.right = _insert(node.right, node)
            return node

        self.root = _insert(self.root)
        self.size += 1

    def remove(self, data: T) -> None:
        """Removes a node in the tree."""

        def _reassign_nodes(
            node: BinaryTreeNode[T], new_child: BinaryTreeNode[T] | None
        ) -> None:
            if new_child is not None:
                new_child.parent = node.parent

            if node.parent is None:
                self.root = new_child
            elif node.parent.right == node:
                node.parent.right = new_child
            else:
                node.parent.left = new_child

        def _get_lowest_node(node: BinaryTreeNode[T]) -> BinaryTreeNode[T]:
            if node.left is None:
                lowest_node = node
                _reassign_nodes(node, node.right)
            else:
                lowest_node = _get_lowest_node(node.left)
            return lowest_node

        node = self.search(data)
        if node is None:
            raise Exception(f"TreeNode with data {data} does not exist")
        self.size -= 1

        # If count is greater than 1, we just decrease the count and return. We reduce
        # the node count regardless in case we are holding a reference to the node.
        node.count -= 1
        if node.count > 0:
            return
        if node.right is None:
            _reassign_nodes(node, node.left)
        elif node.left is None:
            _reassign_nodes(node, node.right)
        else:
            lowest_node = _get_lowest_node(node.right)
            lowest_node.left = node.left
            lowest_node.right = node.right
            if node.left is not None:
                node.left.parent = lowest_node
            if node.right is not None:
                node.right.parent = lowest_node
            _reassign_nodes(node, lowest_node)

    def max_element(self) -> T:
        """Gets the max data inserted in the tree."""
        if self.root is None:
            raise Exception("Binary search tree is empty")
        node = self.root
        while node.right is not None:
            node = node.right
        return node.data

    def min_element(self) -> T:
        """Gets the min data inserted in the tree."""
        if self.root is None:
            raise Exception("Binary search tree is empty")
        node = self.root
        while node.left is not None:
            node = node.left
        return node.data

    def traverse(self, method: str = "inorder") -> Iterator[T]:
        """Return the pre-order, in-order, or post-order traversal of the tree."""
        if method not in ("preorder", "inorder", "postorder"):
            raise ValueError(
                "Method must be one of: 'preorder', 'inorder', or 'postorder'"
            )

        def _traverse(node: BinaryTreeNode[T] | None) -> Iterator[T]:
            if node is not None:
                if method == "preorder":
                    yield node.data
                yield from _traverse(node.left)
                if method == "inorder":
                    yield node.data
                yield from _traverse(node.right)
                if method == "postorder":
                    yield node.data

        return _traverse(self.root)

    def select(self, rank: int) -> T:
        """
        Takes in an integer rank and returns the rank-th order statistic.

        (Equivalent code, but this solution avoids evaluating the iterator)
            return list(self)[rank].data
        """
        return next(node for i, node in enumerate(self) if i == rank).data

    def rank_of(self, data: T) -> int:
        return next(i for i, node in enumerate(self) if node.data == data)



from collections.abc import Sequence
from typing import TypeVar

from cs.util import Comparable

T = TypeVar("T", bound=Comparable)


def binary_search(arr: Sequence[T], target: T) -> int:
    """
    Returns the index of target element, or -1 if it cannot be found.
    Performs a left binary search, which is equivalent to:
        bisect.bisect_left(arr, target)

    Runtime: O(log n)
    """
    left = 0
    right = len(arr) - 1
    while left <= right:
        mid = left + (right - left) // 2
        mid_val = arr[mid]
        if target == mid_val:
            return mid
        if mid_val < target:
            left = mid + 1
        else:
            right = mid - 1
    return -1


def left_right_binary_search(arr: Sequence[T], target: T, is_left: bool = True) -> int:
    """
    Returns the leftmost index of target element, or -1 if it cannot be found.
    Returns rightmost index if is_left is False. Allows duplicates.
    Equivalent to:
        bisect.bisect_left(arr, target)
        bisect.bisect_right(arr, target) - 1

    Runtime: O(log n)
    """
    left = 0
    right = len(arr) - 1
    index = -1
    while left <= right:
        mid = left + (right - left) // 2
        mid_val = arr[mid]
        if mid_val == target:
            index = mid
            if is_left:
                right = mid - 1
            else:
                left = mid + 1
        elif mid_val < target:
            left = mid + 1
        else:
            right = mid - 1
    return index


def linear_search(arr: Sequence[T], target: T) -> int:
    """
    Returns the index of target element, or -1 if it cannot be found.

    Runtime: O(n)
    """
    for i, item in enumerate(arr):
        if item == target:
            return i
    return -1


def binary_search_recursive(arr: Sequence[T], target: T) -> int:
    if not arr:
        return -1
    midpoint = len(arr) // 2
    if arr[midpoint] == target:
        return midpoint
    if target < arr[midpoint]:
        return binary_search(arr[:midpoint], target)
    return binary_search(arr[midpoint + 1 :], target)



"""
A Hamiltonian cycle (Hamiltonian circuit) is a graph cycle through a graph that visits
each node exactly once. Determining whether such paths and cycles exist in graphs is
the 'Hamiltonian path problem', which is NP-complete. Wikipedia:
https://en.wikipedia.org/wiki/Hamiltonian_path
"""
from __future__ import annotations

from typing import cast

from cs.structures import Graph, V


def hamiltonian_cycle(graph: Graph[V], start: V) -> list[V]:
    """
    Either return array of vertices indicating the hamiltonian cycle
    or an empty list indicating that hamiltonian cycle was not found.
    """

    def hamilton_cycle(graph: Graph[V], path: list[V | None], curr_ind: int) -> bool:
        prev = path[curr_ind - 1]
        if prev is None:
            raise RuntimeError

        if curr_ind == len(graph):
            # Return whether path exists between current and starting vertices
            return start in graph[prev]

        for neighbor in graph[prev]:
            if neighbor not in path:
                path[curr_ind] = neighbor
                if hamilton_cycle(graph, path, curr_ind + 1):
                    return True
                path[curr_ind] = None
        return False

    path: list[V | None] = [None] * (len(graph) + 1)
    path[0] = path[-1] = start
    return cast(list[V], path) if hamilton_cycle(graph, path, 1) else []



"""
Functions useful for doing molecular chemistry:
* molarity_to_normality
* moles_to_pressure
* moles_to_volume
* pressure_and_volume_to_temperature
"""


def molarity_to_normality(nfactor: int, moles: float, volume: float) -> float:
    """
    Convert molarity to normality.
    Volume is taken in litres.

    Wikipedia reference: https://en.wikipedia.org/wiki/Equivalent_concentration
    Wikipedia reference: https://en.wikipedia.org/wiki/Molar_concentration
    """
    return round(moles / volume * nfactor)


def moles_to_pressure(volume: float, moles: float, temperature: float) -> float:
    """
    Convert moles to pressure.
    Ideal gas laws are used.
    Temperature is taken in kelvin.
    Volume is taken in litres.
    Pressure has atm as SI unit.

    Wikipedia reference: https://en.wikipedia.org/wiki/Gas_laws
    Wikipedia reference: https://en.wikipedia.org/wiki/Pressure
    Wikipedia reference: https://en.wikipedia.org/wiki/Temperature
    """
    return round(moles * 0.0821 * temperature / volume)


def moles_to_volume(pressure: float, moles: float, temperature: float) -> float:
    """
    Convert moles to volume.
    Ideal gas laws are used.
    Temperature is taken in kelvin.
    Volume is taken in litres.
    Pressure has atm as SI unit.

    Wikipedia reference: https://en.wikipedia.org/wiki/Gas_laws
    Wikipedia reference: https://en.wikipedia.org/wiki/Pressure
    Wikipedia reference: https://en.wikipedia.org/wiki/Temperature
    """
    return round(moles * 0.0821 * temperature / pressure)


def pressure_and_volume_to_temperature(
    pressure: float, moles: float, volume: float
) -> float:
    """
    Convert pressure and volume to temperature.
    Ideal gas laws are used.
    Temperature is taken in kelvin.
    Volume is taken in litres.
    Pressure has atm as SI unit.

    Wikipedia reference: https://en.wikipedia.org/wiki/Gas_laws
    Wikipedia reference: https://en.wikipedia.org/wiki/Pressure
    Wikipedia reference: https://en.wikipedia.org/wiki/Temperature
    """
    return round(pressure * volume / (0.0821 * moles))



class RMQ:
    def __init__(self, elems: list[int]) -> None:
        self.elems = elems

    def rmq(self, low: int, high: int) -> int:
        raise NotImplementedError

    def return_smaller_index(self, index1: int, index2: int) -> int:
        if -1 in (index1, index2):
            return index1
        if self.elems[index1] == self.elems[index2]:
            return index1 if index1 < index2 else index2
        return index1 if self.elems[index1] < self.elems[index2] else index2



def radix_sort(array: list[int]) -> list[int]:
    """
    Radix sort algorithm implementation.

    Let k be the number of digits:
    Runtime: O(nk)
    """
    if not array:
        return array
    max_element = max(array)
    place = 1
    while max_element // place > 0:
        # Apply counting sort to sort elements based on place value.
        output = [0] * len(array)
        count = [0] * 10

        # Calculate count of elements
        for item in array:
            count[(item // place) % 10] += 1

        # Calculate cummulative count
        for i in range(1, 10):
            count[i] += count[i - 1]

        # Place the elements in sorted order
        i = len(array) - 1
        while i >= 0:
            index = array[i] // place
            output[count[index % 10] - 1] = array[i]
            count[index % 10] -= 1
            i -= 1

        # Move to the next place
        array = output
        place *= 10
    return array



"""
https://en.wikipedia.org/wiki/Burrows%E2%80%93Wheeler_transform

The Burrows–Wheeler transform (BWT, also called block-sorting compression)
rearranges a character string into runs of similar characters. This is useful
for compression, since it tends to be easy to compress a string that has runs
of repeated characters by techniques such as move-to-front transform and
run-length encoding. More importantly, the transformation is reversible,
without needing to store any additional data except the position of the first
original character. The BWT is thus a "free" method of improving the efficiency
of text compression algorithms, costing only some extra computation.
"""


def all_rotations(s: str) -> list[str]:
    """
    :param s: The string that will be rotated len(s) times.
    :return: A list with the rotations.
    """
    return [s[i:] + s[:i] for i in range(len(s))]


def bwt_transform(s: str) -> tuple[str, int]:
    """
    :param s: The string that will be used at bwt algorithm
    :return: the string composed of the last char of each row of the ordered
    rotations and the index of the original string at ordered rotations list
    """
    rotations = sorted(all_rotations(s))
    return "".join([word[-1] for word in rotations]), rotations.index(s)


def reverse_bwt(bwt_string: str, idx_original_string: int) -> str:
    """
    :param bwt_string: The string returned from bwt algorithm execution
    :param idx_original_string: A 0-based index of the string that was used to
    generate bwt_string at ordered rotations list
    :return: The string used to generate bwt_string when bwt was executed
    """
    if not bwt_string or not 0 <= idx_original_string < len(bwt_string):
        raise RuntimeError("Invalid input to reverse_bwt")

    ordered_rotations = [""] * len(bwt_string)
    for _ in range(len(bwt_string)):
        for i, bwt in enumerate(bwt_string):
            ordered_rotations[i] = bwt + ordered_rotations[i]
        ordered_rotations.sort()
    return ordered_rotations[idx_original_string]



from __future__ import annotations

from dataclasses import dataclass, field


@dataclass(init=False)
class SuffixArray:
    text: str
    suffix_arr: list[int] = field(repr=False)

    def __init__(self, text: str) -> None:
        from cs.algorithms.string.sais import build_suffix_array

        self.text = text + "$"
        self.suffix_arr = build_suffix_array(text)

    def __str__(self) -> str:
        return str([self.text[suffix_index:] for suffix_index in self.suffix_arr])

    def __getitem__(self, key: int) -> int:
        return self.suffix_arr[key]

    def search(self, pattern: str) -> list[int]:
        """
        Returns a list of indices where all matching strings start.
        Uses a two-pass binary search.
        """

        def left_right_binary_search(target: str, is_left: bool) -> int | None:
            """
            Returns the leftmost index of target element,
            or -1 if it cannot be found.
            """
            left, right = 0, len(self.text) - 1
            index = None
            while left <= right:
                mid = left + (right - left) // 2
                mid_val = self.text[self.suffix_arr[mid] :]
                if mid_val.startswith(target):
                    index = mid
                    if is_left:
                        right = mid - 1
                    else:
                        left = mid + 1
                elif mid_val < target:
                    left = mid + 1
                else:
                    right = mid - 1
            return index

        low = left_right_binary_search(pattern, True)
        high = left_right_binary_search(pattern, False)
        if low is None or high is None:
            return []
        return sorted(self.suffix_arr[x] for x in range(low, high + 1))



from __future__ import annotations

from collections import deque

from cs.structures import Graph, V


def breadth_first_search(graph: Graph[V], start: V, end: V) -> list[V]:
    """
    Identical to DFS except with a queue and pop(0).
    Does not benefit from an additional visited check because it uses a queue.

    Runtime: O(V + E)
    """
    queue: deque[tuple[V, list[V]]] = deque([(start, [start])])
    visited: set[V] = set()
    while queue:
        vertex, path = queue.popleft()
        if vertex == end:
            return path
        visited.add(vertex)
        for neighbor, edge in graph[vertex].items():
            if neighbor not in visited and edge.weight > 0:
                queue.append((neighbor, path + [neighbor]))
    return []



from __future__ import annotations

from collections.abc import Callable, Iterator
from dataclasses import dataclass
from typing import Generic, TypeVar

from cs.util import Comparable

T = TypeVar("T", bound=Comparable)


@dataclass(init=False)
class BinaryHeap(Generic[T]):
    """
    A generic Heap class, represented using an array. We use an array
    rather than a binary tree because heaps grow in a balanced manner, which
    means that the data fits in contiguous memory and no array space is wasted.

    Can be used as min or max heap by passing the according key function.
    """

    _heap: list[T]

    def __init__(self, *, key: Callable[[T], T] = lambda x: x) -> None:
        self._heap = []
        self.elem_to_index: dict[T, int] = {}
        self.size = 0
        self.key = key

    def __len__(self) -> int:
        return self.size

    def __bool__(self) -> bool:
        return bool(self._heap)

    def __contains__(self, item: T) -> bool:
        return item in self.elem_to_index

    def __iter__(self) -> Iterator[T]:
        yield from self._heap

    def __getitem__(self, index: int) -> T:
        if not 0 <= index < self.size:
            raise KeyError(f"Invalid index: {index}")
        return self._heap[index]

    def update(self, item: T, new_item: T) -> None:
        """Updates given item value in heap if present."""
        if item not in self.elem_to_index:
            raise KeyError("Item not found")
        index = self.elem_to_index[item]
        self._heap[index] = new_item
        self.elem_to_index[new_item] = index

        # Make sure heap is right in both up and down direction.
        # Ideally only one of them will make any change.
        self._heapify_up(index)
        self._heapify_down(index)

    def dequeue(self, item: T) -> None:
        """Deletes given item from heap if present."""
        if item not in self.elem_to_index:
            raise KeyError("Item not found")
        index = self.elem_to_index[item]
        del self.elem_to_index[item]
        self._heap[index] = self._heap[self.size - 1]
        self.elem_to_index[self._heap[self.size - 1]] = index
        self.size -= 1
        # Make sure heap is right in both up and down direction. Ideally, only one
        # of them will make any change, so no performance loss in calling both.
        if self.size > index:
            self._heapify_up(index)
            self._heapify_down(index)

    def enqueue(self, item: T) -> None:
        """Inserts given item with given value in heap."""
        new_node = item
        if len(self._heap) == self.size:
            self._heap.append(new_node)
        else:
            self._heap[self.size] = new_node
        self.elem_to_index[item] = self.size
        self.size += 1
        self._heapify_up(self.size - 1)

    def peek(self) -> T:
        """Returns top item tuple (Calculated value, item) from heap if present."""
        if self.size == 0:
            raise ValueError("Heap is empty.")
        return self._heap[0]

    def pop(self) -> T:
        """
        Return top item tuple (Calculated value, item) from heap and removes it as well
        if present.
        """
        top_item = self.peek()
        self.dequeue(top_item)
        return top_item

    def _parent(self, i: int) -> int | None:
        """Returns parent index of given index if exists, else None."""
        return ((i - 1) // 2) if 0 < i < self.size else None

    def _left(self, i: int) -> int | None:
        """Returns left-child-index of given index if exists, else None."""
        left = int(2 * i + 1)
        return left if 0 < left < self.size else None

    def _right(self, i: int) -> int | None:
        """Returns right-child-index of given index if exists, else None."""
        right = int(2 * i + 2)
        return right if 0 < right < self.size else None

    def _swap(self, i: int, j: int) -> None:
        """Performs changes required for swapping two elements in the heap."""
        # First update the indexes of the items in index map.
        self.elem_to_index[self._heap[i]], self.elem_to_index[self._heap[j]] = (
            self.elem_to_index[self._heap[j]],
            self.elem_to_index[self._heap[i]],
        )
        # Then swap the items in the list.
        self._heap[i], self._heap[j] = self._heap[j], self._heap[i]

    def _cmp(self, i: int, j: int) -> bool:
        """Compares the two items using default comparison."""
        return self.key(self._heap[i]) < self.key(self._heap[j])

    def _get_valid_parent(self, i: int) -> int:
        """
        Returns index of valid parent as per desired ordering among given index and
        both its children.
        """
        left, right = self._left(i), self._right(i)
        valid_parent = i

        if left is not None and not self._cmp(left, valid_parent):
            valid_parent = left
        if right is not None and not self._cmp(right, valid_parent):
            valid_parent = right

        return valid_parent

    def _heapify_up(self, index: int) -> None:
        """Fixes the heap in upward direction of given index."""
        parent = self._parent(index)
        while parent is not None and not self._cmp(index, parent):
            self._swap(index, parent)
            index, parent = parent, self._parent(parent)

    def _heapify_down(self, index: int) -> None:
        """Fixes the heap in downward direction of given index."""
        valid_parent = self._get_valid_parent(index)
        while valid_parent != index:
            self._swap(index, valid_parent)
            index, valid_parent = valid_parent, self._get_valid_parent(valid_parent)



# pylint: disable=too-many-branches
from __future__ import annotations

import math
from collections import deque
from dataclasses import dataclass
from typing import Generic, TypeVar
from uuid import UUID, uuid4

from cs.structures.heap.heap import Heap
from cs.util import Comparable, formatter

T = TypeVar("T", bound=Comparable)


@dataclass(order=True)
class Entry(Generic[T]):
    """
    Hold an entry in the heap.
    In order for all of the Fibonacci heap operations to complete in O(1), clients need
    to have O(1) access to any element in the heap. We make this work by having each
    insertion operation produce a handle to the node in the tree. In actuality, this
    handle is the node itself.

    Priority is the first parameter because the dataclass orders Entry as a tuple
    (priority, value).

    Note that the input type should be bounded by Comparable in order for the
    implementation to type-check. Otherwise, the heap should still work so long as there
    are no equal priorities. If you want to use a non-comparable type with this class,
    use a wrapper class with a custom comparator or only use unique priorities.
    """

    priority: float
    value: T

    def __post_init__(self) -> None:
        """Initialize an Entry in the heap."""
        # Number of children
        self.degree = 0
        self.is_marked = False
        self.parent: Entry[T] | None = None
        self.child: Entry[T] | None = None
        self.next = self.prev = self

    def __repr__(self) -> str:
        return str(formatter.pformat(self))


@dataclass(init=False)
class FibonacciHeap(Heap[T]):
    """See docs/fibonacci_heap.md for code credits and implementation details."""

    top: Entry[T] | None

    def __init__(self, *, allow_duplicates: bool = False) -> None:
        # Pointer to the minimum element in the heap.
        self.top: Entry[T] | None = None

        # Mapping from element to corresponding entry.
        # Should not introduce any asymptotic change in memory overhead.
        self.elem_to_entry: dict[T | UUID, Entry[T]] = {}

        # Cached size of the heap, so we don't have to recompute this explicitly.
        self.size = 0

        # Whether to allow duplicate key entries, which means using UUIDs for
        # elem_to_entry's keys instead.
        self.allow_duplicates = allow_duplicates

    def __bool__(self) -> bool:
        return self.top is not None

    def __len__(self) -> int:
        return self.size

    def __contains__(self, item: T) -> bool:
        if self.allow_duplicates:
            raise NotImplementedError
        return item in self.elem_to_entry

    def __getitem__(self, value: T | UUID) -> Entry[T]:
        """Gets the correct Entry object from the given value or UUID."""
        if self.allow_duplicates and not isinstance(value, UUID):
            raise RuntimeError(
                "You must pass in a valid UUID or set allow_duplicates = False."
            )
        if not self.allow_duplicates and isinstance(value, UUID):
            raise RuntimeError("You must pass in a value of type T, not a UUID.")

        if value not in self.elem_to_entry:
            raise KeyError(
                f"Invalid {'UUID' if self.allow_duplicates else 'key'}: {value}"
            )

        return self.elem_to_entry[value]

    def __or__(self, other: object) -> FibonacciHeap[T]:
        # TODO - this is a destructive operation
        if isinstance(other, FibonacciHeap):
            result = FibonacciHeap[T]()
            self.merge(other)
            result.merge(self)
            return result
        raise NotImplementedError

    def __ior__(self, other: object) -> None:
        if isinstance(other, FibonacciHeap):
            self.merge(other)
        raise NotImplementedError

    @staticmethod
    def merge_lists(one: Entry[T] | None, two: Entry[T] | None) -> Entry[T] | None:
        """
        Merge 2 lists.

        Utility function which, given two pointers into disjoint circularly- linked
        lists, merges the two lists together into one circularly-linked list in O(1)
        time. Because the lists may be empty, the return value is the only pointer
        that's guaranteed to be to an element of the resulting list.

        This function assumes one and two are the minimum elements of the lists they are
        in, and returns a pointer to whichever is smaller. If this condition does not
        hold, the return value is some arbitrary pointer into the doubly-linked list.

        @param one A reference to one of the two deques.
        @param two A reference to the other of the two deques.
        @return A reference to the smallest element of the resulting list.
        """
        if one is None:
            return two or None

        if two is None:
            return one or None

        # Both non-None; actually do the splice.
        # We have two lists that look like this and we want to switch A and B.
        # +----+     +----+     +----+
        # |    |--N->|one |--N->|  A |
        # |    |<-P--|    |<-P--|    |
        # +----+     +----+     +----+
        #
        # +----+     +----+     +----+
        # |    |--N->|two |--N->|  B |
        # |    |<-P--|    |<-P--|    |
        # +----+     +----+     +----+
        one.next, two.next = two.next, one.next
        one.next.prev, two.next.prev = one, two
        return one if one < two else two

    @staticmethod
    def _check_priority(priority: float) -> None:
        """
        Given a user-specified priority, check whether it's a valid double
        and throw a ValueError otherwise.

        @param priority The user's specified priority.
        @raises ValueError if it is not valid.
        """
        if math.isnan(priority):
            raise ValueError(f"Priority {priority} is invalid.")

    def enqueue(self, value: T, priority: float = 0) -> T | UUID:
        """
        Insert an element into the Fibonacci heap with the specified priority.

        Its priority must be a valid double, so you cannot set the priority to NaN.

        @param value The value to insert.
        @param priority Its priority, which must be valid.
        @return An Entry representing that element in the tree.
        """
        self._check_priority(priority)
        if not self.allow_duplicates and value in self.elem_to_entry:
            raise KeyError(
                f"Duplicate key detected: {value}. "
                "Use allow_duplicates = True to allow duplicate entries using UUIDs."
            )

        # Create the entry object, which is a circularly-linked list of length one.
        result = Entry(priority, value)

        # Merge this singleton list with the tree list.
        self.top = self.merge_lists(self.top, result)
        self.size += 1

        key: T | UUID = uuid4() if self.allow_duplicates else value
        self.elem_to_entry[key] = result
        return key

    def peek(self) -> tuple[T, float]:
        """
        Return an Entry object corresponding to the minimum element of the heap.

        Raise an IndexError if the heap is empty.

        @return The smallest element of the heap.
        @raises IndexError If the heap is empty.
        """
        if self.top is None:
            raise IndexError("Heap is empty.")
        top = self.top
        return top.value, top.priority

    def dequeue(self) -> tuple[T, float]:
        """
        Dequeue and return the minimum element of the Fibonacci heap.

        If the heap is empty, this throws an IndexError.

        @return The smallest element of the Fibonacci heap.
        @raises IndexError if the heap is empty.
        """
        if self.top is None:
            raise IndexError("Heap is empty.")

        # Otherwise, we're about to lose an element, so decrement the number of
        # entries in this heap.
        self.size -= 1

        # Grab the minimum element so we know what to return.
        min_elem = self.top
        return_val = (min_elem.value, min_elem.priority)

        # Now, we need to get rid of this element from the list of roots. There
        # are two cases to consider. First, if this is the only element in the
        # list of roots, we set the list of roots to be None by clearing top.
        # Otherwise, if it's not None, then we write the elements next to the
        # min element around the min element to remove it, then arbitrarily
        # reassign the min.
        if self.top.next is self.top:
            self.top = None
        else:
            self.top.prev.next = self.top.next
            self.top.next.prev = self.top.prev
            # Arbitrary element of the root list.
            self.top = self.top.next

        # Next, clear the parent fields of all of the min element's children,
        # since they're about to become roots. Because the elements are
        # stored in a circular list, the traversal is a bit complex.
        if min_elem.child is not None:
            # Keep track of the first visited node.
            curr = min_elem.child
            while True:
                curr.parent = None

                # Walk to the next node, then stop if this is the node we
                # started at.
                curr = curr.next
                if curr is min_elem.child:
                    # This was a do-while (curr != minElem.mChild);
                    break

        # Next, splice the children of the root node into the topmost list,
        # then set self.top to point somewhere in that list.
        self.top = self.merge_lists(self.top, min_elem.child)

        # If there are no entries left, we're done.
        if self.top is None:
            if not self.allow_duplicates:
                del self.elem_to_entry[min_elem.value]
            return return_val

        # Next, we need to coalesce all of the roots so that there is only one
        # tree of each degree. To track trees of each size, we allocate an
        # ArrayList where the entry at position i is either None or the
        # unique tree of degree i.
        tree_table: deque[Entry[T] | None] = deque()

        # We need to traverse the entire list, but since we're going to be
        # messing around with it we have to be careful not to break our
        # traversal order mid-stream. One major challenge is how to detect
        # whether we're visiting the same node twice. To do this, we'll
        # spent a bit of overhead adding all of the nodes to a list, and
        # then will visit each element of this list in order.
        to_visit: deque[Entry[T]] = deque()

        # To add everything, we'll iterate across the elements until we
        # find the first element twice. We check this by looping while the
        # list is empty or while the current element isn't the first element
        # of that list.

        curr = self.top
        while not to_visit or to_visit[0] is not curr:
            to_visit.append(curr)
            curr = curr.next

        # Traverse this list and perform the appropriate unioning steps.
        for curr in to_visit:
            # Keep merging until a match arises.
            while True:
                # Ensure that the list is long enough to hold an element of this degree.
                while curr.degree >= len(tree_table):
                    tree_table.append(None)

                # If nothing's here, we can record that this tree has this size
                # and are done processing.
                if tree_table[curr.degree] is None:
                    tree_table[curr.degree] = curr
                    break

                # Otherwise, merge with what's there.
                other = tree_table[curr.degree]
                if other is None:
                    raise RuntimeError

                # Clear the slot
                tree_table[curr.degree] = None

                # Determine which of the two trees has the smaller root, storing
                # the two trees accordingly.
                minimum = other if other < curr else curr
                maximum = curr if other < curr else other

                # Break max out of the root list, then merge it into min's child list.
                maximum.next.prev = maximum.prev
                maximum.prev.next = maximum.next

                # Make it a singleton so that we can merge it.
                maximum.next = maximum.prev = maximum
                minimum.child = self.merge_lists(minimum.child, maximum)

                # Reparent maximum appropriately.
                maximum.parent = minimum

                # Clear maximum's mark, since it can now lose another child.
                maximum.is_marked = False

                # Increase minimum's degree; it now has another child.
                minimum.degree += 1

                # Continue merging this tree.
                curr = minimum

            # Update the global min based on this node. Note that we compare
            # for <= instead of < here. That's because if we just did a
            # reparent operation that merged two different trees of equal
            # priority, we need to make sure that the min pointer points to
            # the root-level one.
            if curr <= self.top:
                self.top = curr
        if not self.allow_duplicates:
            del self.elem_to_entry[min_elem.value]
        return return_val

    def decrease_key(self, value: T | UUID, new_priority: float) -> None:
        """
        Decrease the key of the specified element to the new priority.

        If the new priority is greater than the old priority, this function raises an
        ValueError. The new priority must be a finite double, so you cannot set the
        priority to be NaN, or +/- infinity. Doing so also raises an ValueError.

        @param entry The element whose priority should be decreased.
        @param new_priority The new priority to associate with this entry.
        @raises ValueError If the new priority exceeds the old
                priority, or if the argument is not a finite double.
        """
        entry = self[value]
        self._check_priority(new_priority)
        if new_priority > entry.priority:
            raise ValueError("New priority exceeds old.")
        self._decrease_key_unchecked(entry, new_priority)

    def remove(self, value: T | UUID) -> None:
        """
        Remove this Entry from the Fibonacci heap that contains it.

        @param entry The entry to delete.
        """
        # Use decreaseKey to drop the entry's key to -infinity. This will
        # guarantee that the node is cut and set to the global minimum.
        self._decrease_key_unchecked(self[value], float("-inf"))
        self.dequeue()

    def merge(self, other: Heap[T]) -> None:
        """
        Merge 2 Fibonacci heaps.

        Given two Fibonacci heaps, obtain a Fibonacci heap that contains
        all of the elements of the two heaps in place.

        @param self The first Fibonacci heap to merge.
        @param other The second Fibonacci heap to merge.
        """
        if not isinstance(other, FibonacciHeap):
            raise TypeError("Heap types must match when merging.")

        if not self.allow_duplicates and set(self.elem_to_entry) & set(
            other.elem_to_entry
        ):
            raise RuntimeError(
                "You must pass in two unoverlapping heaps or set "
                "allow_duplicates = True on both heaps."
            )

        # Merge the two Fibonacci heap root lists together. This helper function
        # also computes the min of the two lists, so we can store the result in
        # the top field of the new heap.
        self.top = self.merge_lists(self.top, other.top)

        # The size of the new heap is the sum of the sizes of the input heaps.
        self.size += other.size
        self.allow_duplicates = self.allow_duplicates or other.allow_duplicates
        self.elem_to_entry |= other.elem_to_entry

    def _decrease_key_unchecked(self, entry: Entry[T], priority: float) -> None:
        """
        Decrease the key of a node in the tree without doing any checking to ensure
        that the new priority is valid.

        @param entry The node whose key should be decreased.
        @param priority The node's new priority.
        """
        # First, change the node's priority.
        entry.priority = priority

        # If the node no longer has a higher priority than its parent, cut it.
        # Note that this also means that if we try to run a delete operation
        # that decreases the key to -infinity, it's guaranteed to cut the node
        # from its parent.
        if entry.parent is not None and entry <= entry.parent:
            self._cut_node(entry)

        # If our new value is the new min, mark it as such. Note that if we
        # ended up decreasing the key in a way that ties the current minimum
        # priority, this will change the min accordingly.
        if self.top is not None and entry <= self.top:
            self.top = entry

    def _cut_node(self, entry: Entry[T]) -> None:
        """
        Cut a node from its parent.

        If the parent was already marked, recursively cuts that node from its
        parent as well.

        @param entry The node to cut from its parent.
        """
        # Begin by clearing the node's mark, since we just cut it.
        entry.is_marked = False

        # Base case: If the node has no parent, we're done.
        if entry.parent is None:
            return

        # Rewire the node's siblings around it, if it has any siblings.
        if entry.next is not entry:
            # Has siblings
            entry.next.prev = entry.prev
            entry.prev.next = entry.next

        # If the node is the one identified by its parent as its child,
        # we need to rewrite that pointer to point to some arbitrary other child.
        if entry.parent.child is entry:
            # If there are any other children, pick one of them arbitrarily.
            # Otherwise, there aren't any children left and we should clear the
            # pointer and drop the node's degree.
            entry.parent.child = None if entry.next is entry else entry.next

        # Decrease the degree of the parent, since it just lost a child.
        entry.parent.degree -= 1

        # Splice this tree into the root list by converting it to a singleton
        # and invoking the merge subroutine.
        entry.prev = entry.next = entry
        self.top = self.merge_lists(self.top, entry)

        # Mark the parent and recursively cut it if it's already been
        # marked.
        if entry.parent.is_marked:
            self._cut_node(entry.parent)
        else:
            entry.parent.is_marked = True

        # Clear the relocated node's parent; it's now a root.
        entry.parent = None



from cs.structures import Graph, V


def topological_sort(graph: Graph[V]) -> list[V]:
    """
    Perform topological sort on a directed acyclic graph.

    Node never seen = not in visited
    Node being processed = in visited but not in stack
    Node done = in stack

    Runtime: O(V + E)
    """

    def depth_first_search(v: V, visited: set[V], stack: list[V]) -> None:
        visited.add(v)
        for neighbor in graph[v]:
            if neighbor in visited and neighbor not in stack:
                raise ValueError(f"Cycle detected in node {v}")
            if neighbor not in visited:
                depth_first_search(neighbor, visited, stack)
        stack.append(v)

    stack: list[V] = []
    visited: set[V] = set()
    for v in graph:
        if len(visited) == len(graph):
            break
        if v not in visited:
            depth_first_search(v, visited, stack)
    return stack



"""
See docs/karatsuba.md for code credits and implementation details.
Author: Keith Schwarz (htiek@cs.stanford.edu)
"""


def add(lhs: list[int], rhs: list[int], base: int) -> list[int]:
    """
    Adds two arbitrary-precision values in some base together.

    Given two arrays lhs and rhs of digits in some base 'base,' returns an
    array of the number lhs + rhs encoded in base 'base.'
    """

    # Pad the two inputs to be the same length.
    length = max(len(lhs), len(rhs))
    lhs = [0] * (length - len(lhs)) + lhs
    rhs = [0] * (length - len(rhs)) + rhs

    # Track the carry from the previous column; initially this is zero.
    carry = 0

    # Track the result.  We'll build the array up in reverse to avoid costly
    # prepend operations that aren't relevant.
    result = []

    # Iterate across the digits in reverse, computing the sum.
    for i in range(1, len(lhs) + 1):
        # Sum the carry and the two values in this column
        column = lhs[-i] + rhs[-i] + carry

        # Output the column value (after modding by the base)
        result.append(column % base)

        # Update the carry
        carry = column // base

    # Prepend the carry to the result if it's nonzero.
    if carry != 0:
        result.append(carry)

    # Reverse the order of the resulting digits; that's the more proper way
    # to hand them back.
    result.reverse()

    return result


def subtract(lhs: list[int], rhs: list[int], base: int) -> list[int]:
    """
    Subtracts two arbitrary-precision values in some base.

    Given two arrays lhs and rhs of digits in some base 'base,' returns an
    array of the number lhs - rhs encoded in base 'base.'  It is assumed that
    lhs >= rhs; an error occurs if this is not the case.
    """

    # Pad the two inputs to be the same length.
    length = max(len(lhs), len(rhs))
    lhs = [0] * (length - len(lhs)) + lhs
    rhs = [0] * (length - len(rhs)) + rhs

    # Track the result.  We'll build the array up in reverse to avoid costly
    # prepend operations that aren't relevant.
    result = []

    # Iterate across the digits in reverse, subtracting the values.
    for i in range(1, len(lhs) + 1):
        # Compute the difference in this column.
        difference = lhs[-i] - rhs[-i]

        # If we can subtract without borrowing, do so.
        if difference >= 0:
            result.append(difference)
            continue

        # Otherwise, we have to borrow from previous columns.
        for j in range(i + 1, length + 1):
            # Drop the value of this digit by one.  We do this by adding
            # base - 1 and then modding by base, which is equivalent to
            # subtracting one modulo base.
            lhs[-j] = (lhs[-j] + (base - 1)) % base

            # If this value didn't wrap around (i.e. its value is not
            # base - 1), then we're done.  Otherwise, we have to go to the
            # previous digit.
            if lhs[-j] != base - 1:
                break

        # Finally, update the result by appending the correct value.
        # Since we just computed lhs - rhs, we actually want to compute
        # (base + lhs) - rhs.
        result.append(difference + base)

    # Reverse the order of the resulting digits; that's the more proper way
    # to hand them back.
    result.reverse()

    return result


def multiply(lhs: list[int], rhs: list[int], base: int) -> list[int]:
    """
    Multiplies two arbitrary-precision values in some base.

    Given two arrays of lhs and rhs of digits in some base 'base,' returns
    an array of digits corresponding to their product using the Karatsuba
    algorithm.
    """
    assert len(lhs) > 0
    assert len(rhs) > 0

    # Pad the two inputs to be the same length.
    length = max(len(lhs), len(rhs))
    lhs = [0] * (length - len(lhs)) + lhs
    rhs = [0] * (length - len(rhs)) + rhs

    # If the numbers are one digit each, just multiply them and convert the
    # answer back to an (up to) two digit number.
    if length == 1:
        # Compute the true answer.
        result = lhs[0] * rhs[0]

        # Convert it back to an array.
        return [result] if result < base else [result // base, result % base]

    # Otherwise, we need to use Karatsuba's recursive algorithm to compute the
    # values.  To do this, we'll first compute how many digits we'll put into
    # each of the smaller numbers.  This is given by ceil(length / 2), which
    # can be represented beautifully by computing (length + 1) / 2.  This
    # works because if length is even (length + 1) / 2 = (2n + 1) / 2 = n
    # when using integer division, and if length is odd (length + 1) / 2 =
    # (2n + 1 + 1) / 2 = (2n + 2) / 2 = n + 1.
    m0 = (length + 1) // 2
    m1 = length // 2

    # Split the inputs in half.
    x0, x1 = lhs[:m0], lhs[m0:]
    y0, y1 = rhs[:m0], rhs[m0:]

    # Compute p0, p1, and p2.
    p0 = multiply(x0, y0, base)
    p1 = multiply(add(x0, x1, base), add(y0, y1, base), base)
    p2 = multiply(x1, y1, base)

    # Since z0 = p0 and z2 = p2, we don't need to compute them.  However, we
    # do need to compute z1 = p1 - p0 - p2.
    z0 = p0
    z1 = subtract(p1, add(p0, p2, base), base)
    z2 = p2

    # From these results, compute z0 b^(2m) + z1 b^m + z2.  We separate out
    # each of these operations.
    z0prod = z0 + [0] * (2 * m1)
    z1prod = z1 + [0] * m1
    z2prod = z2

    return add(add(z0prod, z1prod, base), z2prod, base)


def karatsuba(a: int, b: int) -> int:
    """
    Alternative karatsuba implementation for integers instead of lists.
    This is similar to the actual Python implempentation of multiplication.
    """
    len_a, len_b = len(str(a)), len(str(b))
    if len_a == 1 or len_b == 1:
        return a * b

    m1 = max(len_a, len_b)
    m2 = m1 // 2

    a1, a2 = divmod(a, 10**m2)
    b1, b2 = divmod(b, 10**m2)

    x = karatsuba(a2, b2)
    y = karatsuba(a1 + a2, b1 + b2)
    z = karatsuba(a1, b1)

    result = (z * 10 ** (2 * m2)) + ((y - z - x) * 10**m2) + x
    return int(result)
